/*
 * Copyright (C) 2017 CERN (www.cern.ch)
 * Author: Tristan Gingold <tristan.gingold@cern.ch>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 *
 * test_vme: program to test the vme64xcore (to be used with the
 * svec_vmecore_test design).
 */

#define _XOPEN_SOURCE 600
#define _GNU_SOURCE 1
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdarg.h>
#include <unistd.h>
#include <getopt.h>
#include "libvmebus.h"

#define VME_2eVME 3

static enum vme_dma_block_size vbs = VME_DMA_BSIZE_2048;
static enum vme_dma_block_size pbs = VME_DMA_BSIZE_2048;

static void
rate_start (struct timespec *start_ts)
{
  if (clock_gettime (CLOCK_MONOTONIC, start_ts) != 0)
    {
      fprintf (stderr, "clock_gettime error: %m\n");
      exit (4);
    }
}

static void
rate_done (struct timespec *start_ts, unsigned int nbr_bytes)
{
  struct timespec end_ts;
  unsigned long nano;

  if (clock_gettime (CLOCK_MONOTONIC, &end_ts) != 0)
    {
      fprintf (stderr, "clock_gettime error: %m\n");
      exit (4);
    }
  nano = (end_ts.tv_nsec - start_ts->tv_nsec)
    + (end_ts.tv_sec - start_ts->tv_sec) * 1000000000;
  printf ("Rate: %f MB/sec\n",
	  nbr_bytes * 1.0E9 / nano / (1 << 20));
}

static enum vme_dma_block_size
convert_block_size (const char *str)
{
  unsigned int num;

  num = strtoul (str, NULL, 0);
  switch (num)
    {
    case 32:
      return VME_DMA_BSIZE_32;
    case 64:
      return VME_DMA_BSIZE_64;
    case 128:
      return VME_DMA_BSIZE_128;
    case 256:
      return VME_DMA_BSIZE_256;
    case 512:
      return VME_DMA_BSIZE_512;
    case 1024:
      return VME_DMA_BSIZE_1024;
    case 2048:
      return VME_DMA_BSIZE_2048;
    case 4096:
      return VME_DMA_BSIZE_4096;
    default:
      fprintf (stderr, "unhandled block size of %u\n", num);
      exit (2);
    }
}

static void
do_test_dma (uint8_t am, enum vme_2esst_mode vmode,
             uint32_t addr, uint32_t len, void *buf)
{
  struct vme_dma dma;
  int s;
  struct timespec start_ts;

  memset (&dma, 0, sizeof (dma));
  dma.status = 0; // ??
  dma.length = len;
  dma.novmeinc = 0;
  dma.dir = VME_DMA_FROM_DEVICE;

  dma.src.data_width = VME_D32;
  dma.src.am = am;
  dma.src.v2esst_mode = vmode;
  dma.src.bcast_select = 0;
  dma.src.addru = 0;
  dma.src.addrl = addr;

  dma.dst.addru = ((uintptr_t)buf) >> 32;
  dma.dst.addrl = (uintptr_t)buf;

  dma.ctrl.vme_block_size = vbs;
  dma.ctrl.vme_backoff_time = VME_DMA_BACKOFF_0;
  dma.ctrl.pci_block_size = pbs;
  dma.ctrl.pci_backoff_time = VME_DMA_BACKOFF_0;

  rate_start (&start_ts);
  s = vme_dma_read (&dma);
  rate_done (&start_ts, len);

  if (s != 0)
    printf ("Error: DMA read\n");
}

#define OPT_VBS 128
#define OPT_PBS 129

static const struct option options[] =
  {
   { "addr", required_argument, 0, 'a' },
   { "len",  required_argument, 0, 'n' },
   { "am",   required_argument, 0, 'm' },
   { "2e",   required_argument, 0, 'e' },
   { "dump", no_argument,       0, 'd' },
   { "help", no_argument,       0, 'h' },
   { "vbs",  required_argument, 0, OPT_VBS },
   { "pbs",  required_argument, 0, OPT_PBS },
   { NULL, 0, 0, 0 }
  };

int
main (int argc, char **argv)
{
  int am = 0x08; /* MBLT */
  unsigned addr = 0xffffffff;
  unsigned int len = 4096;
  int c;
  void *buf;
  int flag_dump = 0;
  enum vme_2esst_mode vmode = VME_2eVME;

  while ((c = getopt_long (argc, argv, "a:m:n:e:dh",
                           options, NULL)) != -1)
    switch (c)
      {
      case 'a':
	addr = strtoul (optarg, NULL, 0);
	break;
      case 'n':
	len = strtol (optarg, NULL, 0);
	break;
      case 'm':
	am = strtol (optarg, NULL, 0);
	break;
      case 'e':
        if (strcmp (optarg, "2eVME") == 0)
          vmode = VME_2eVME;
        else if (strcmp (optarg, "2eSST") == 0
                 || strcmp (optarg, "2eSST-160") == 0)
          vmode = VME_SST160;
        else {
          fprintf (stderr, "%s: bad value for -e, expect 2eVME or 2eSST-160\n",
                   argv[0]);
          exit(3);
        }
        break;
      case 'd':
        flag_dump = 1;
        break;
      case OPT_VBS:
        vbs = convert_block_size (optarg);
        break;
      case OPT_PBS:
        pbs = convert_block_size (optarg);
        break;
      case 'h':
	printf ("usage: %s -a ADDR [-m AM] [-n LENGTH]\n", argv[0]);
        printf ("options:\n"
                " --vbs=SZ  set VME block size\n"
                " --pbs=SZ  set PCI block size\n");
	return 0;
      case '?':
	fprintf (stderr, "incorrect option, try %s -h\n", argv[0]);
	return 1;
      }

  if (addr == 0xffffffff)
    {
      fprintf (stderr, "missing address\n");
      return 2;
    }
  if (am == 0)
    {
      fprintf (stderr, "bad AM\n");
      return 2;
    }

  printf ("INFO: test DMA for am 0x%02x, mode: %u (addr: 0x%08x, len: 0x%0x)\n",
          am, vmode, addr, len);

  if (posix_memalign (&buf, 4096, len) != 0)
    {
      fprintf (stderr, "cannot allocate memory\n");
      exit (4);
    }

  do_test_dma (am, vmode, addr, len, buf);

  if (flag_dump)
    for (unsigned i = 0; i < len; i += 4)
      printf ("%08x: %08x\n", addr + i, ((unsigned int *)buf)[i >> 2]);

  free (buf);

  return 0;
}
